/******************/
/* GameBoy -> Z80 */
/******************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	MAXLINE		255
#define	MAXDATA		0x10000

int main(int argc, char **argv)
{
	FILE	*infp, *outfp, *cmdfp;
	char	line[MAXLINE];
	int		patchdata[MAXDATA];
	int		stadd,st,ed;
	int		cin,pin;
	int		outdt;
	int		cmdnum;

	int		cmddat[256] = {
		1,3,1,1,1,1,2,1,	/* 0x00-0x07 */
		3,1,1,1,1,1,2,1,	/* 0x08-0x0f */
		2,3,1,1,1,1,2,1,	/* 0x10-0x17 */
		2,1,1,1,1,1,2,1,	/* 0x18-0x1f */
		2,3,1,1,1,1,2,1,	/* 0x20-0x27 */
		2,1,1,1,1,1,2,1,	/* 0x28-0x2f */
		2,3,1,1,1,1,2,1,	/* 0x30-0x37 */
		2,1,1,1,1,1,2,1,	/* 0x38-0x3f */
		1,1,1,1,1,1,1,1,	/* 0x40-0x47 */
		1,1,1,1,1,1,1,1,	/* 0x48-0x4f */
		1,1,1,1,1,1,1,1,	/* 0x50-0x57 */
		1,1,1,1,1,1,1,1,	/* 0x58-0x5f */
		1,1,1,1,1,1,1,1,	/* 0x60-0x67 */
		1,1,1,1,1,1,1,1,	/* 0x68-0x6f */
		1,1,1,1,1,1,1,1,	/* 0x70-0x77 */
		1,1,1,1,1,1,1,1,	/* 0x78-0x7f */
		1,1,1,1,1,1,1,1,	/* 0x80-0x87 */
		1,1,1,1,1,1,1,1,	/* 0x88-0x8f */
		1,1,1,1,1,1,1,1,	/* 0x90-0x97 */
		1,1,1,1,1,1,1,1,	/* 0x98-0x9f */
		1,1,1,1,1,1,1,1,	/* 0xa0-0xa7 */
		1,1,1,1,1,1,1,1,	/* 0xa8-0xaf */
		1,1,1,1,1,1,1,1,	/* 0xb0-0xb7 */
		1,1,1,1,1,1,1,1,	/* 0xb8-0xbf */
		1,1,3,3,3,1,2,1,	/* 0xc0-0xc7 */
		1,1,3,2,3,3,2,1,	/* 0xc8-0xcf */
		1,1,3,0,3,1,2,1,	/* 0xd0-0xd7 */
		1,1,3,0,3,0,2,1,	/* 0xd8-0xdf */
		2,1,1,0,0,1,2,1,	/* 0xe0-0xe7 */
		2,1,3,0,0,0,2,1,	/* 0xe8-0xef */
		2,1,1,1,0,1,2,1,	/* 0xf0-0xf7 */
		2,1,3,1,0,0,2,1		/* 0xf8-0xff */
	};

	int		i;

	if (argc != 4) {
		fprintf(stderr, "Usage: gb2z80 binfile outfile commandfile\n");
		exit(1);
	}

/* open files */

	if ((infp = fopen(argv[1], "rb")) == NULL) {
		perror(argv[1]);
		exit(1);
	}

	if ((outfp = fopen(argv[2], "wb")) == NULL) {
		perror(argv[2]);
		exit(1);
	}

	if ((cmdfp = fopen(argv[3], "r")) == NULL) {
		perror(argv[3]);
		exit(1);
	}

	for (i = 0;i < MAXDATA;i ++) {
		patchdata[i] = 0;
	}

	if ( fgets(line, MAXLINE, cmdfp) != NULL) {
		sscanf(line,"%x",&stadd);
	}

	while ( fgets(line, MAXLINE, cmdfp) != NULL) {
		sscanf(line,"%x-%x",&st,&ed);
		for (i = st-stadd;i <= ed-stadd;i ++) {
			patchdata[i] = 1;
		}
	}

	for (i = 0;;i ++) {
		cin = fgetc(infp);
		if (cin == EOF) {
			break;
		}

		if (patchdata[i] == 0) {
			outdt = cin;
			cmdnum = 0;
		} else {
			if (cmdnum == 0) {
				cmdnum = cmddat[cin];
				if (cmdnum == 0) {
					fprintf(stderr,"Illigal address = %04x : data = %02x\n",i,cin);
					cmdnum = 1;
				}
				switch (cin) {
					case 0xea:			/* LD (nn),A */
						outdt = 0x32;
						break;
					case 0xfa:			/* LD A,(nn) */
						outdt = 0x3a;
						break;
					case 0xe0:			/* LD ($ffxx),A */
						outdt = 0xc7;	/* RST $00 */
						break;
					case 0xf0:			/* LD A,($ffxx) */
						outdt = 0xcf;	/* RST $08 */
						break;
					case 0x2a:			/* LD A,(HL+) */
						outdt = 0xe7;	/* RST $20 */
						break;
					case 0x22:			/* LD (HL+),A */
						outdt = 0xef;	/* RST $28 */
						break;
					case 0xd9:			/* RETI -> RET */
						outdt = 0xc9;
						break;
					default:
						outdt = cin;
						break;
				}

			} else {
				outdt = cin;
			}
			cmdnum --;
		}

		fputc( outdt, outfp );
	}


	fclose(infp);
	fclose(outfp);
	fclose(cmdfp);


	return(0);
}
